import multiprocessing
import os

class counter:
	def __init__ (self, initial_value=0):
		self.value = initial_value
	
	def inc(self, delta=1):
		return self.increment(delta)
	
	def dec(self, delta=1):
		return self.decrement(delta)	
		
	def increment (self, delta=1):
		result = self.value
		try:
			self.value = self.value + delta
		except OverflowError:
			self.value = int(self.value) + delta
		return result

	def decrement (self, delta=1):
		result = self.value
		try:
			self.value = self.value - delta
		except OverflowError:
			self.value = int(self.value) - delta
		return result

	def as_long (self):
		return int (self.value)

	def __bool__ (self):
		return self.value != 0

	def __repr__ (self):
		return '<counter value=%s at %x>' % (self.value, id(self))

	def __str__ (self):
		return str(int(self.value))

class Value:
	def __init__ (self, type, initial_value):
		self.value = initial_value
	
class mpcounter:
	def __init__ (self, initial_value=0):
		if os.environ.get ("SKITAI_ENV") == "PYTEST":
			import threading
			self.value = Value ("i", initial_value)
			self.lock = threading.Lock ()			
		else:
			self.value = multiprocessing.Value ("i", initial_value)
			self.lock = multiprocessing.Lock ()
		
	def inc(self, delta=1):
		return self.increment(delta)
	
	def dec(self, delta=1):
		return self.decrement(delta)	
		
	def increment (self, delta=1):
		self.lock.acquire ()
		try:
			result = self.value.value
			try:
				self.value.value = self.value.value + delta
			except OverflowError:
				self.value.value = int(self.value.value) + delta
		finally:
			self.lock.release ()					
		return result

	def decrement (self, delta=1):
		self.lock.acquire ()
		try:
			result = self.value.value
			try:
				self.value.value = self.value.value - delta
			except OverflowError:
				self.value.value = int(self.value.value) - delta
		finally:
			self.lock.release ()		
		return result

	def as_long (self):
		self.lock.acquire ()
		result = self.value.value
		self.lock.release ()
		return int (result)

	def __bool__ (self):
		self.lock.acquire ()
		result = self.value.value
		self.lock.release ()
		return result != 0

	def __repr__ (self):
		self.lock.acquire ()
		result = self.value.value
		self.lock.release ()
		return '<mpcounter value=%s at %x>' % (result, id(self))

	def __str__ (self):
		self.lock.acquire ()
		result = self.value.value
		self.lock.release ()
		return str(int(result))

if __name__ == "__main__":
	f = mpcounter (0)
	print(f.increment ())
	print(f.increment ())
	print(f.increment ())
	print(f.as_long ())
	print(str (f))
	print(f.decrement ())
	print(f.as_long ())
	print(str (f))
	
	
	




import multiprocessing
import rs4
from concurrent.futures import TimeoutError
from rs4.logger import screen_logger
import time

N_CPU = multiprocessing.cpu_count()

class Future:
    def __init__ (self, future, func):
        self.future = future
        self._name = "{}.{}".format (func.__module__, func.__name__)
        self._started = time.time ()

    def then (self, r):
        return r

    def __str__ (self):
        return self._name

    def __getattr__ (self, name):
        return getattr (self.future, name)


class Executor:
    def __init__ (self, executor_class, workers = N_CPU, logger = None):
        self._executor_class = executor_class 
        self._name = self._executor_class.__name__       
        self.logger = logger
        self.workers = workers
        self.lock = multiprocessing.Lock ()
        self.executor = None
        self.futures = []

    def __len__ (self):
        with self.lock:
            return len (self.futures)

    def maintern (self):
        with self.lock:
            inprogresses = []
            for future in self.futures:
                if not future.done ():
                     inprogresses.append (future)       
                     continue
                expt = future.exception (0)
                if expt:
                    try:
                        raise expt
                    except:
                        self.logger.trace (future._name)
            self.futures = inprogresses

    def create_excutor (self):
        self.executor = self._executor_class (self.workers)        

    def shutdown (self):
        if not self.executor:
            return

        with self.lock:
            for future in self.futures:
                if not future.done ():
                    continue
                if not future.running ():
                    future.cancel ()
                    continue
                try:    
                    future.result (timeout = 3.0)
                except TimeoutError:
                    pass

        self.executor.shutdown (wait = True)
        self.maintern ()
        return len (self.futures)

    def __call__ (self, f, *a, **b):     
        if self.executor is None:
            self.create_excutor ()
        else:
            self.maintern ()
        future = self.executor.submit (f, *a, **b)
        wrap = Future (future, f)
        self.logger ("started {}: {}".format (self._name, wrap))
        with self.lock:            
            self.futures.append (wrap)        
        return wrap

class Executors:
    def __init__ (self, workers = N_CPU, logger = None):
        self.logger = logger or screen_logger ()
        self.executors = [
            Executor (rs4.threading, workers, self.logger),
            Executor (rs4.processing, workers, self.logger)
        ]        
    
    def status (self):
        return dict (
            threads = len (self.executors [0]),
            processes = len (self.executors [1])
        )

    def cleanup (self):        
        return [e.shutdown () for e in self.executors]
    
    def create_thread (self, f, *a, **b):
        return self.executors [0] (f, *a, **b)

    def create_process (self, f, *a, **b):
        return self.executors [1] (f, *a, **b)

        


import asyncore
import time

def _make (was, flt):
	g = {}
	
	if flt:
		if flt.startswith ("CLUSTER:"):
			return {flt: was.clusters [flt[8:]].status ()}
		elif flt == "SOCKETPOOL":
			return {flt: was.clusters [None].status ()}
		elif flt == "ENVIRON":
			return {"ENVIRON": was.env}	
	else:
		for name, cluster in list(was.clusters.items ()):
			if name is None: 
				g ["SOCKETPOOL"] = cluster.status ()
				continue
			g ["CLUSTER:" + name] = cluster.status ()
			g ["ENVIRON"] = was.env
			
	for attr in dir (was):
		if attr.startswith ("__") or attr == "config": continue
		if flt and flt != attr.upper (): continue
		obj = eval ("was." + attr)
		if hasattr (obj, "status"):			
			g [attr.upper ()] = obj.status ()
	
	return g
	
	
def format_object (o, depth = 0):
	b = []
	if type (o) in (type ([]), type (())):
		for each in o:
			if type (each) is type ({}):
				b.append ("%s" % format_object (each, depth + 1))
			elif type (each) is type ([]):
				b.append (", ".join ([str (x) for x in each]))				
			else:					
				if not b:
					b.append ("<ul>")
				b.append ("<li>%s</li>" % str (each))
		if b and b [0] == "<ul>":
			b.append ("</ul>")		
		return "".join (b)

	b.append ("<table width='100%' border='0'>")
	ii = list(o.items ())
	ii.sort ()
	width = 18 + (depth * 6)
	for idx, (k1, v1) in enumerate (ii):
		if isinstance (v1, (list, dict)):
			b.append ("<tr><td bgcolor='#C8EFD4' valign='top' width='%d%%' style='word-break: break-all; padding: 6px;'><b>%s%s</b></td>" % (width, str (k1), type (v1) in (list, tuple) and " (%d)" % (len (v1),) or ""))			
			b.append ("<td valign='top' style='padding: 0;'>%s</td></tr>" % format_object (v1, depth + 1))			
		else:
			if depth > 0 and idx == 0:
				b.append ("<tr><td valign='top' colspan='2' style='padding: 6; color: #888; font-size: 11px;'><b>Object</b></td></tr>")		
			b.append ("<tr><td bgcolor='#98E0AD' valign='top' width='%d%%' style='word-break: break-all; padding: 6px;'><b>%s</b></td>" % (width, str (k1)))
			b.append ("<td valign='top' style='word-break: break-all; padding: 6px;'>%s</td></tr>" % str (v1))
	b.append ("</table>")	
	return "".join (b)

	
def formatting (was, info, flt):	
	b = [
		"<!DOCTYPE html>"
		"<html><head>"
		"<title>Skitai App Engine Status</title>"
		"</head><body style='font-size: 12px;''>"
	]
	l = list(info.items ())
	l.sort ()
	if not flt:
		b.append ('<a name="TOC"></a><h3>Table of Content</h3>')
		b.append ('<ol>')
		got_cluster = False
		for k, v in l:
			if k.startswith ("CLUSTER:") and not got_cluster:
				got_cluster = True
				b.append ("<li>CLUSTERS</li>")
				b.append ('<ul style="list-style: square;">')
			elif got_cluster and not k.startswith ("CLUSTER:"):
				got_cluster = False
				b.append ('</ul>')
			b.append ("<li><a href='#%s'>%s</a></li>" % (k, got_cluster and k [8:] or k))
		b.append ('</ol>')
	
	for k, v in l:
		if not flt:
			b.append ("<a name='%s'></a><h3><a href='%s?f=%s'>%s</a></h3>" % (k, was.request.split_uri ()[0], k, k))
		else:
			b.append ("<h3><a href='%s'>%s</a></h3>" % (was.request.split_uri ()[0], k))	
			
		b.append (format_object (v))
		if not flt:
			b.append ('<div style="margin-bottom: 30px; padding: 5px;"><a href="#TOC">TOC</a></div>')
		else:
			b.append ('<div style="margin-bottom: 30px; padding: 5px;"><a href="%s">TOC</a></div>' % was.request.split_uri ()[0])
	b.append ("</body></html>")		
	return "".join (b)
	

def make (was, flt = None, fancy = True):
	info = _make (was, flt)
	if fancy:
		return formatting (was, info, flt)	
	return info



import os, sys
import time
import tempfile
from hmac import new as hmac
import copy
from hashlib import sha1
import json
import xmlrpc.client as xmlrpclib
import base64
import pickle
import random
import threading
from rs4 import pathtool, logger, jwt, deco
from aquests.protocols.smtp import composer
from aquests.protocols.http import http_date, util
from skitai import __version__, WS_EVT_OPEN, WS_EVT_CLOSE, WS_EVT_INIT, NAME
from skitai import lifetime
from . import server_info
from ..corequest import tasks
from .. import http_response
from .promise import Promise
from .triple_logger import Logger
if os.environ.get ("SKITAI_ENV") == "PYTEST":
    from threading import RLock    
else:    
    from multiprocessing import RLock

import xmlrpc.client as xmlrpclib
from rs4.producers import file_producer
from .api import DateEncoder

if os.environ.get ("SKITAI_ENV") == "PYTEST":
    from .semaps import TestSemaps as Semaps
else:    
    from .semaps import Semaps

if os.name == "nt":
    TEMP_DIR =  os.path.join (tempfile.gettempdir(), "skitai-gentemp")
else:
    TEMP_DIR = "/var/tmp/skitai-gentemp"        
pathtool.mkdir (TEMP_DIR)


class WASBase:
    version = __version__    
    objects = {}
    _luwatcher = Semaps ([], "d", 256)    
    lock = plock = RLock ()
    init_time = time.time ()    
    cloned = False    
    _process_locks = [RLock () for i in range (8)]
    
    @classmethod    
    def get_lock (cls, name = "__main__"):
        return cls._process_locks [hash (name) % 8]
    
    # application friendly methods -----------------------------------------
    @classmethod
    def register (cls, name, obj):
        if hasattr (cls, name):
            raise AttributeError ("server object `%s` is already exists" % name)
        cls.objects [name] = obj
        setattr (cls, name, obj)
    
    @classmethod
    def unregister (cls, name):
        del cls.objects [name]
        return delattr (cls, name)
        
    @classmethod
    def add_handler (cls, back, handler, *args, **karg):
        h = handler (cls, *args, **karg)
        if hasattr (cls, "httpserver"):
            cls.httpserver.install_handler (h, back)
        return h
    
    def _clone (self, if_origin = False):
        if self.cloned:
            return self    
        new_was = copy.copy (self)
        new_was.cloned = True
        return new_was        
    
    @property
    def timestamp (self):
        return int (time.time () * 1000) 
    
    @property
    def uniqid (self):
        return "{}{}".format (self.timestamp, self.gentemp () [-7:])
        
    def __dir__ (self):
        return self.objects.keys ()
    
    def __str__ (self):
        return "skitai was for {}".format (threading.currentThread ())
    
    def in__dict__ (self, name):
        return name in self.__dict__
    
    # utils -----------------------------------------------
    def txnid (self):
        return "%s/%s" % (self.request.gtxid, self.request.ltxid)
    
    def rebuild_header (self, header, method, data = None, internal = True):
        nheader = util.normheader (header)
        if method in {"get", "delete", "post", "put", "patch", "upload"}:
            try:
                default_request_type = self.app.config.get ("default_request_type")
            except AttributeError:
                default_request_type = None
            if not default_request_type:    
                default_request_type = self.DEFAULT_REQUEST_TYPE            
            util.set_content_types (nheader, data, default_request_type)
            
        if internal:
            nheader ["X-Gtxn-Id"] = self.request.get_gtxid ()
            nheader ["X-Ltxn-Id"] = self.request.get_ltxid (1)
        else:
            nheader ["X-Requested-With"] = NAME
        return nheader
    
    # system functions ----------------------------------------------    
    def Tasks (self, reqs, timeout = 10, **args):
        self.response.set_timeout (timeout)
        return tasks.Tasks (reqs, timeout, **args)

    def Mask (self, data):
        return tasks.Mask (data)

    def Thread (self, func, *args, **kargs):
        return self.executors.create_thread (func, *args, **kargs)

    def Process (self, func, *args, **kargs):
        return self.executors.create_process (func, *args, **kargs)

    def log (self, msg, category = "info", at = "app"):
        self.logger (at, msg, "%s:%s" % (category, self.txnid ()))
        
    def traceback (self, id = "", at = "app"):
        if not id:
            id = self.txnid ()
        self.logger.trace (at, id)
    
    def render_ei (self, exc_info, format = 0):
        return http_response.catch (format, exc_info)
    
    @property
    def tempfile (self):
        return self.gentemp () 
    
    def gentemp (self):
        return os.path.join (TEMP_DIR, next (tempfile._get_candidate_names()))
    
    def email (self, subject, snd, rcpt):
        if composer.Composer.SAVE_PATH is None:
            composer.Composer.SAVE_PATH = os.path.join ("/var/tmp/skitai", "smtpda", "spool")
            pathtool.mkdir (composer.Composer.SAVE_PATH)
        return composer.Composer (subject, snd, rcpt)
    
    # -- only allpy to current worker process
    def status (self, flt = None, fancy = True):        
        return server_info.make (self, flt, fancy)
    
    def restart (self, timeout = 0):
        lifetime.shutdown (3, timeout)
    
    def shutdown (self, timeout = 0):
        lifetime.shutdown (0, timeout)
    
    # inter-processes communication ------------------------------    
    def setlu (self, name, *args, **karg):
        self._luwatcher.set (name, time.time (), karg.get ("x_ignore", False))
        self.broadcast (name, *args, **karg)            
        
    def getlu (self, *names):
        mtimes = []
        for name in names:
            mtime = self._luwatcher.get (name, self.init_time)
            mtimes.append (mtime)
        return max (mtimes)
    
    def setgs (self, name, val):
        assert isinstance (val, int), "global state must be integer"
        self._luwatcher.set (name, val)
        self.broadcast (name, val)
        
    def getgs (self, name, default = 0):
        assert isinstance (default, int), "global state must be integer"
        val = self._luwatcher.get (name, default)
        return int (val)

    # websokcet / http2 ------------------------------------------    
    def push (self, uri):
        self.request.response.hint_promise (uri)
        
    def wsconfig (self, spec, timeout = 60, encoding = "text"):
        # other than atila, encoding became varname
        self.env ["websocket.config"] = (spec, timeout, encoding)
        return ""
        
    def wsinit (self):
        return self.env.get ('websocket.event') == WS_EVT_INIT
    
    def wsopened (self):
        return self.env.get ('websocket.event') == WS_EVT_OPEN
    
    def wsclosed (self):
        return self.env.get ('websocket.event') == WS_EVT_CLOSE
    
    def wshasevent (self):
        return self.env.get ('websocket.event')
    
    def wsclient (self):
        return self.env.get ('websocket.client')
    
    # will be deprecated --------------------------------------------------
    def togrpc (self, obj):
        return obj.SerializeToString ()
    
    def fromgrpc (self, message, obj):
        return message.ParseFromString (obj)
        
    def tojson (self, obj):
        return json.dumps (obj, cls = DateEncoder)
        
    def toxml (self, obj):
        return xmlrpclib.dumps (obj, methodresponse = False, allow_none = True, encoding = "utf8")    
    
    def fromjson (self, obj):
        if type (obj) is bytes:
            obj = obj.decode ('utf8')
        return json.loads (obj)
    
    def fromxml (self, obj, use_datetime = 0):
        return xmlrpclib.loads (obj)
    
    def fstream (self, path, mimetype = 'application/octet-stream'):    
        self.response.set_header ('Content-Type',  mimetype)
        self.response.set_header ('Content-Length', str (os.path.getsize (path)))    
        return file_producer (open (path, "rb"))
            
    def jstream (self, obj, key = None):
        self.response.set_header ("Content-Type", "application/json")
        if key:
            # for single skeleton data is not dict
            return self.tojson ({key: obj})
        else:
            return self.tojson (obj)        
    
    def xstream (self, obj, use_datetime = 0):            
        self.response.set_header ("Content-Type", "text/xml")
        return self.toxml (obj, use_datetime)
    
    def gstream (self, obj):
        self.response.set_header ("Content-Type", "application/grpc")
        return self.togrpc (obj)
    